# encoding: utf-8
# frozen_string_literal: true

module Carbon
  module Concrete
    module Item
      # A function definition.  This contains all of the information needed to
      # build a proper LLVM function.  This is stored within an index and
      # serialized as needed, in order to defer compilation.  There are two
      # main types of functions: "normal," or "extern."  "Normal" functions
      # have an actual definition (i.e. {Tacky::Function}), and are written
      # in Ruby/Carbon.  "Extern" functions do not have a definition; they
      # map a Carbon function to a C/low-level function.  Instead of a
      # definition, they have a function name, that maps to the C/low-level
      # function.  This is normally specified with an `:extern` directive.
      #
      # @api private
      # @note
      #   **This class is frozen upon initialization.**  This means that any
      #   attempt to modify it will result in an error.  In most cases, the
      #   attributes on this class will also be frozen, as well.
      class Function
        include Base
        attr_reader :return
        attr_reader :definition

        # (see Item::Base.from)
        def self.from(type)
          {
            type: type,
            generics: type.generics,
            internal: type.function.name,
            parameters: type.function.parameters,
            definition: Tacky::Function.new(type.function.parameters)
          }
        end

        def initialize(data)
          @type = data.fetch(:type)
          @internal = data.fetch(:internal) { @type.function.name }
          @parameters = data.fetch(:parameters) { @type.function.parameters }
          @definition = data.fetch(:definition)
          @generics = data.fetch(:generics) { @type.generics }
          @return = data.fetch(:return)
          @extern = data[:extern]
          @name = @type.to_s

          derive_dependencies
          deep_freeze!
        end

        def define(index)
          if @definition.is_a?(::Proc)
            Function.new(Function.from(@type)
              .merge(definition: @definition.call(index), return: @return))
          else
            self
          end
        end

        # (see Base#call)
        def call(build, generics)
          case @definition
          when Tacky::Function
            build_function_intern(build, generics)
          when ::String, ::Symbol
            build_function_extern(build, generics)
          else fail ArgumentError,
            "Unknown definition #{@definition.class}"
          end
        end

      private

        def build_function_intern(build, generics)
          full = @type.sub(generics)
          params = @parameters
            .map { |p| build.fetch(p.sub(generics)).last }
          ret = build.fetch(@return.sub(generics)).last
          func = build.module.functions.add(@extern || full, params, ret)
          @definition.call(func, build, generics)
          build.items[full] = [self, func]
        end

        def build_function_extern(build, generics)
          full = @type.sub(generics)
          params = @parameters
            .map { |p| build.fetch(p.sub(generics)).last }
          ret = build.fetch(@return.sub(generics)).last
          func = build.module.functions.add(@definition, params, ret)
          build.items[full] = [self, func]
        end
        # rubocop:enable Metrics/AbcSize

        def derive_dependencies
          @dependencies = Set.new
          @dependencies << @return
          @dependencies.merge(@parameters)
          @dependencies.merge(@definition.dependencies) if \
            @definition.is_a?(Tacky::Function)
        end
      end
    end
  end
end
