# encoding: utf-8
# frozen_string_literal: true

module Carbon
  module Concrete
    class Type
      # The "name" of a type.  This contains all of the physical information
      # of a type.  This mostly means the "parts" for the name, and the
      # function information if it is included.
      #
      # @note
      #   **This class is frozen upon initialization.**  This means that any
      #   attempt to modify it will result in an error.  In most cases, the
      #   attributes on this class will also be frozen, as well.
      class Name
        include Enumerable
        extend Forwardable

        # The parts of the name.  This is mainly the sections of the type
        # name; for example, for `Carbon::Pointer`, the parts are `Carbon` and
        # `Pointer`.  Each part can contain generic information; however,
        # by convention, only the last part is actually included for generic
        # information.
        #
        # @api semipublic
        # @example
        #   type = Carbon::Type("Carbon::Pointer<T>")
        #   name = type.name
        #   name.parts
        #     # => [#<Carbon::Concrete::Type::Part Carbon>,
        #     #   #<Carbon::Concrete::Type::Part Pointer>]
        # @return [<Type::Part>] The parts.
        attr_reader :parts

        # The function information of a type.  If no function information was
        # included (i.e. the type is a module), then this is `nil`.
        # Function information always includes a name and the parameters, if
        # it is included.
        #
        # @api public
        # @example
        #   full = "Carbon::Pointer<T>.+(Carbon::Pointer<T>, Carbon::Int32)"
        #   type = Carbon::Type(full)
        #   name = type.name
        #   name.function.to_s # => "+(Carbon::Pointer<T>, Carbon::Int32)"
        # @return [Type::Function] The function information.
        attr_reader :function

        # @!method each
        #   Iterates over each part.
        #
        #   @see #parts
        #   @api semipublic
        #   @example
        #     type.name.each { |part| puts part }
        #   @return [::Enumerable] An enumerable over the parts.
        # @!method to_a
        #   Returns an array of the parts of the name.
        #
        #   @see #parts
        #   @api semipublic
        #   @example
        #     type.name.to_a == type.name.parts # => true
        #   @return [<Type::Part>] The parts.
        # @!method [](index)
        #   Returns a part at a specific index, or the parts over the given
        #   range.
        #
        #   @see #parts
        #   @api semipublic
        #   @example
        #     type.name[0].to_s # => "Carbon"
        #   @param index [Numeric, Range<Numeric>]
        #   @return [Type::Part] The part.
        #   @return [<Type::Part>] The parts over the range.
        # @!method first
        #   Returns the first part in the name.
        #
        #   @see #parts
        #   @api semipublic
        #   @example
        #     type.name.first.to_s # => "Carbon"
        #   @return [Type::Part] The first part.
        # @!method last
        #   Returns the last part in the name.
        #
        #   @see #part
        #   @api semipublic
        #   @example
        #     type.name.last.to_s # => "Pointer<T>"
        #   @return [Type::Part] The last part.
        delegate [:each, :to_a, :[], :first, :last] => :@parts

        # Initialize the name with the given parts and function information.
        # Freezes the class after completion.
        #
        # @param parts [<Type::Part>] The parts of the name.
        # @param function [Type::Function?] The function part of the name, if
        #   it exists.
        def initialize(parts, function)
          @parts = parts.freeze
          @function = function

          to_s
          intern
          deep_freeze!
        end

        # A string representation of the name.  Unlike {#intern}, this
        # includes all generic information as well.
        #
        # @api public
        # @example
        #   type = Carbon::Type("Carbon::Pointer<T>")
        #   type.to_s # => "Carbon::Pointer<T>"
        # @return [::String] The string representation.
        def to_s
          @_to_s ||=
            if @function
              "#{@parts.map(&:to_s).join('::')}.#{@function}"
            else
              @parts.map(&:to_s).join("::")
            end.freeze
        end

        # An interned name of the type.  Unlike {#to_s}, this doesn't include
        # any generic information.
        #
        # @api public
        # @example
        #   type = Carbon::Type("Carbon::Pointer<T>")
        #   type.intern # => "Carbon::Pointer"
        # @note See {Item::Base#intern} For more information about interned
        #   names.
        # @return [::String] The interned name of the type.
        def intern
          @_intern ||=
            if @function
              "#{@parts.map(&:intern).join('::')}.#{@function.intern}"
            else
              @parts.map(&:intern).join("::")
            end.freeze
        end
      end
    end
  end
end
