# encoding: utf-8
# frozen_string_literal: true
# rubocop:disable all

require "strscan"

module Carbon
  module Concrete
    class Type
      # Parses a type.  This starts by scanning the string, then parsing it.
      # Each instance should result in a different parsed value; a single
      # instance cannot be re-used for parsing something else.  **Not for
      # public consumption.**
      #
      # @api private
      # @private
      class Parse
        # Initialize the parser.
        #
        # @param scan [::String] The string to parse.
        def initialize(scan)
          @string = scan
          @scanner = StringScanner.new(@string)
        end

        # The resulting value.  This should be a name.
        #
        # @see #parse
        # @return [Type::Name]
        def value
          @_value ||= parse
        end

      private

        def parse
          @tokens = scan
          name = parse_name
        end

        def scan
          return to_enum(:scan) unless block_given?

          until @scanner.eos?
            single = scan_single
            yield single if single
          end

          yield [:EOS]
        end

        def expect(*values)
          token = @tokens.next
          error(values, token[0]) unless values.include?(token[0])
          token
        end

        def parse_name
          parts = [parse_name_part]
          func = nil

          while @tokens.peek[0] == :"::"
            expect :"::"
            parts << parse_name_part
          end

          parts = [*parts[0..-2].map(&:strip), parts[-1]]
          func = parse_function_name if @tokens.peek[0] == :"."

          Name.new(parts, func)
        end

        def parse_function_name
          name = nil
          params = []
          expect :"."
          name =  case @tokens.peek[0]
                  when :<   then expect(:<); "<"
                  when :>   then expect(:>); ">"
                  when :+   then expect(:+); "+"
                  else expect(:FNAME)[1]
                  end

          expect :"("
          until @tokens.peek[0] == :")"
            params << Type.new(parse_name)
            break unless @tokens.peek[0] == :","
            expect :","
          end
          expect :")"

          Function.new(name, params)
        end

        def parse_name_part
          part = expect(:MNAME)
          generics = []
          if @tokens.peek[0] == :<
            expect :<

            until @tokens.peek[0] == :>
              generics << parse_generic
              break unless @tokens.peek[0] == :","
              expect :","
            end

            expect :>
          end

          Part.new(part[1], generics)
        end

        def parse_generic
          name = Type.new(parse_name)
          implements = []
          if @tokens.peek[0] == :":"
            expect :":"
            implements << Type.new(parse_name)
            while @tokens.peek[0] == :+
              expect :+
              implements << Type.new(parse_name)
            end
          end

          Generic.new(name, implements)
        end

        FUNC_NAME = %r{
          (
              [a-z][A-Za-z0-9_-]*[!?=]?
            | === | ==
            | \<=\>
            | << | >>
            | <= | >=
            | \-
            | \* | \/
            | % | & | \|
            | \[\]=
            | \[\]
            | \^
          )
        }x.freeze

        def scan_single
          case
          when @scanner.scan(/[A-Z][A-Za-z0-9_-]*/)
            [:MNAME, @scanner[0]]
          when @scanner.scan(FUNC_NAME)
            [:FNAME, @scanner[0]]
          when @scanner.scan(/::/) then [:"::"]
          when @scanner.scan(/</) then [:<]
          when @scanner.scan(/>/) then [:>]
          when @scanner.scan(/,/) then [:","]
          when @scanner.scan(/:/) then [:":"]
          when @scanner.scan(/\+/) then [:+]
          when @scanner.scan(/\./) then [:"."]
          when @scanner.scan(/\(/) then [:"("]
          when @scanner.scan(/\)/) then [:")"]
          when @scanner.scan(/\s+/)
          else error([], @string[@scanner.pos])
          end
        end

        def error(expected, given)
          fail ArgumentError,
            "Unexpected #{given.inspect}, " \
            "expected #{expected.map(&:inspect).join(', ')} " \
            "(source #{@string})"
        end
      end
    end
  end
end
