# encoding: utf-8
# frozen_string_literal: true

module Carbon
  module Core
    # An integer type in Carbon.  The integer types are kept in here in order
    # to keep track of the properties of each integer type.
    #
    # @api private
    class Int
      # The name of the integer type.
      #
      # @api public
      # @example
      #   int.name # => #<Carbon::Concrete::Type Carbon::Boolean>
      # @return [Carbon::Concrete::Type] The name.
      attr_reader :name

      # The name of the function that casts another integer to this one.
      #
      # @api public
      # @example
      #   int.cast # => "to-bool"
      # @return [::String] The cast function name.
      attr_reader :cast

      # The signage of the integer type.  This is either `:signed` or
      # `:unsigned`.  `:signed` implies two's compliment.
      #
      # @api public
      # @example
      #   int.sign # => :unsigned
      # @return [::Symbol] How the integer is signed.
      attr_reader :sign

      # The number of bits in the integer.  This is normally a power of two.
      #
      # @api public
      # @example
      #   int.size # => 1
      # @return [::Integer] The size of the integer.
      attr_reader :size

      # Finds a specific integer with the given size and signage.  Uses
      # {Ints} as the base.
      #
      # @param size [::Integer] The size of the integer.
      # @param sign [::Symbol] How the integer should be signed.
      # @return [Int] If an integer with the given size and signage is found.
      # @return [nil] Otherwise.
      def self.find(size:, sign:)
        Ints.find { |i| i.size == size && i.sign == sign }
      end

      # Initialize the int.
      #
      # @param name [Carbon::Concrete::Type] The type of the integer.
      # @param cast [::String] The cast function name.
      # @param sign [::Symbol] How the integer is signed.
      # @param size [::Integer] The size of the integer, in bits.
      def initialize(name, cast, sign, size)
        @name = name
        @cast = cast
        @sign = sign
        @size = size

        deep_freeze!
      end
    end

    # The integeral types that are defined in Carbon, and all information
    # pertaining to them.
    #
    # @return [<Int>]
    Ints = [
      ["Carbon::Int64", "to-int64", :signed, 64],
      ["Carbon::UInt64", "to-uint64", :unsigned, 64],
      ["Carbon::Int32", "to-int32", :signed, 32],
      ["Carbon::UInt32", "to-uint32", :unsigned, 32],
      ["Carbon::Int16", "to-int16", :signed, 16],
      ["Carbon::UInt16", "to-uint16", :unsigned, 16],
      ["Carbon::Int8", "to-int8", :signed, 8],
      ["Carbon::UInt8", "to-uint8", :unsigned, 8],
      ["Carbon::Boolean", "to-bool", :unsigned, 1]
    ].map { |a| Int.new(Carbon::Type(a[0]), *a[1..-1]) }.deep_freeze!
  end
end
