# encoding: utf-8
# frozen_string_literal: true

require "carbon/tacky/instruction/generation"

module Carbon
  module Tacky
    # An instruction.  This is the building block of the rest of the Tacky
    # structures.  Each instruction performs some sort of action.
    class Instruction < Value
      include Instruction::Generation
      # Instructions that can be represented by this class that can have a
      # name parameter.
      #
      # @return [<::String>]
      NAMED_INSTRUCTIONS = %w(
        add alloca and array_alloca array_malloc ashr bit_cast exact_sdiv
        extract_element extract_value fadd fcmp fdiv fmul fp2si fp2ui fp_cast
        fp_ext fp_trunc frem fsub gep global_string global_string_pointer icmp
        inbounds_gep insert_element insert_value int2ptr int_cast invoke
        is_not_null is_null load lshr malloc mul neg not nsw_add nsw_mul
        nsw_neg nsw_sub nuw_add nuw_mul nuw_neg nuw_sub or phi pointer_cast
        ptr2int ptr_diff sdiv select sext sext_or_bit_cast shl shuffle_vector
        si2fp srem struct_gep sub trunc trunc_or_bit_cast udiv ui2fp urem xor
        zext zext_or_bit_cast
      ).freeze

      # The name of the instruction.  This is limited to a specific set; see
      # the method names on {Tacky::Builder} for what they can be.
      #
      # @return [::String]
      attr_reader :instruction

      # The parameters of the instruction.
      #
      # @return [<Reference, Parameter, Value, ::Object>]
      attr_reader :parameters

      # Initializes the instruction with the given id, name, and parameters.
      #
      # @param id [::Numeric] The ID of the instruction.  This is the number
      #   of the instruction in the block.
      # @param name [::Symbol, ::String] The name of the instruction.
      # @param parameters [<Reference, Parameter, Value, ::Object>] The
      #   parameters to be used with the instruction.
      def initialize(id, name, parameters, reference)
        @value = id
        @instruction = name.to_s
        @parameters = parameters
        @reference = reference
      end

      # Builds the instruction.  If the instruction is special, i.e. it starts
      # with an underscore, it is handled seperately.  Otherwise, it is handled
      # normally.
      #
      # @api private
      # @see #generate_null
      # @see #generate_sizeof
      # @see #generate_normal
      # @param context [Tacky::Context] The context.
      # @param block [::LLVM::BasicBlock] The block.
      # @return [LLVM::Value]
      def call(context, block)
        case @instruction
        when "_null"   then generate_null(context, block)
        when "_sizeof" then generate_sizeof(context, block)
        when "_deref"  then generate_deref(context, block)
        else generate_normal(context, block)
        end
      end

      # The dependencies of the instruction.  If any of the parameters are a
      # {Concrete::Type}, it is set as a dependency, and the result is put into
      # a set and returned.
      #
      # @api private
      # @return [Set<Concrete::Type>]
      def dependencies
        @parameters.select { |p| p.is_a?(Concrete::Type) }.inject(Set.new, :<<)
      end

      # Retrieves the name of the instruction.  The name of the instruction
      # is essentally the name used by the return value that represents the
      # instruction.  If this instruction is not a named instruction (i.e.
      # {#instruction} is not in {NAMED_INSTRUCTIONS}), or the last parameter
      # isn't a string, it returns an empty string (`""`).  Otherwise, it
      # returns the last parameter.
      #
      # @return [::String] The name of the instruction.
      def name
        if NAMED_INSTRUCTIONS.include?(@instruction) &&
           @parameters.last.is_a?(::String)
          @parameters.last
        else
          ""
        end
      end

      # Sets the name of the instruction.  The name of the instruction is
      # essentially the name used by the return value that represents the
      # instruction.  If this instruction is not a named instruction (i.e.
      # {#instruction} is not in {NAMED_INSTRUCTIONS}), then this does nothing.
      # If the last parameter is a string, this sets the last parameter to the
      # new name.  Otherwise, it adds the name to the end of the parameters
      # list.
      #
      # @param name [::String] The new name of the instruction.
      # @return [::String] The new name of the instruction.
      def name=(name)
        return name unless NAMED_INSTRUCTIONS.include?(@instruction)
        if @parameters.last.is_a?(::String)
          @parameters[-1] = name
        else
          @parameters << name
        end

        name
      end
    end
  end
end
